/*
** pork_perl_xs.c - Perl scripting support
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <config.h>

#include <unistd.h>
#include <sys/types.h>

#include <stdio.h>
#include <stdlib.h>
#include <ncurses.h>
#include <sys/mman.h>

#undef PACKAGE
#undef instr
#define HAS_BOOL

#include <EXTERN.h>
#include <perl.h>
#include <XSUB.h>

#include <libfaim/aim.h>
#include <libfaim/faimconfig.h>

#include <pork.h>
#include <pork_missing.h>
#include <pork_util.h>
#include <pork_list.h>
#include <pork_events.h>
#include <pork_buddy.h>
#include <pork_set.h>
#include <pork_swindow.h>
#include <pork_imwindow.h>
#include <pork_acct.h>
#include <pork_input.h>
#include <pork_bind.h>
#include <pork_screen.h>
#include <pork_screen_cmd.h>
#include <pork_aim.h>
#include <pork_status.h>
#include <pork_screen.h>
#include <pork_timer.h>
#include <pork_command.h>
#include <pork_command_defs.h>
#include <pork_slist.h>
#include <pork_buddy_list.h>
#include <pork_conf.h>
#include <pork_alias.h>
#include <pork_events.h>
#include <pork_perl.h>
#include <pork_perl_xs.h>

extern struct screen screen;

XS(PORK_connect) {
	char *username = NULL;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 0)
		username = SvPV(ST(0), notused);

	cmd_connect(username);
	XSRETURN_IV(0);
}

XS(PORK_disconnect) {
	struct aim_priv *priv;
	dXSARGS;
	dlist_t *node;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	if (!priv->can_connect)
		XSRETURN_IV(-1);

	node = pork_acct_find(priv->username);
	if (node == NULL)
		XSRETURN_IV(-1);

	pork_acct_del(node);
	XSRETURN_IV(0);
}

XS(PORK_err_msg) {
	size_t notused;
	char *msg;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	msg = SvPV(ST(0), notused);
	if (msg == NULL)
		XSRETURN_IV(-1);

	screen_err_msg("%s", msg);
	XSRETURN_IV(0);
}

XS(PORK_status_msg) {
	size_t notused;
	char *msg;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	msg = SvPV(ST(0), notused);
	if (msg == NULL)
		XSRETURN_IV(-1);

	screen_status_msg("%s", msg);
	XSRETURN_IV(0);
}

XS(PORK_load) {
	int args = 0;
	dXSARGS;

	(void) cv;

	if (items < 1)
		XSRETURN_IV(-1);

	while (items > 0) {
		char *file;
		size_t notused;

		file = SvPV(ST(args), notused);
		if (file == NULL)
			XSRETURN_IV(-1);

		if (read_conf(file) != 0)
			XSRETURN_IV(-1);

		args++;
		items--;
	}

	XSRETURN_IV(0);
}

XS(PORK_load_perl) {
	int args = 0;
	dXSARGS;

	(void) cv;

	if (items < 1)
		XSRETURN_IV(-1);

	while (items > 0) {
		char *file;
		size_t notused;

		file = SvPV(ST(args), notused);
		if (file == NULL)
			XSRETURN_IV(-1);

		if (perl_load_file(file) != 0)
			XSRETURN_IV(-1);

		args++;
		items--;
	}

	XSRETURN_IV(0);
}

XS(PORK_echo) {
	size_t notused;
	char *msg;
	dXSARGS;

	(void) cv;

	if (items < 1)
		XSRETURN_IV(-1);

	msg = SvPV(ST(0), notused);
	screen_win_msg(cur_window(), 0, 0, 1, "%s", msg);
	XSRETURN_IV(0);
}

XS(PORK_quit) {
	int exit_val = 0;
	dXSARGS;

	(void) cv;

	if (items > 1)
		XSRETURN_IV(-1);

	if (items >= 1)
		exit_val = SvIV(ST(0));

	pork_exit(exit_val, NULL);
	XSRETURN_IV(0);
}

XS(PORK_refresh) {
	dXSARGS;

	(void) cv;
	(void) items;

	screen_refresh();
	XSRETURN_IV(0);
}

XS(PORK_save) {
	dXSARGS;

	(void) cv;
	(void) items;

	XSRETURN_IV(save_global_config());
}

XS(PORK_bind) {
	char *key_str;
	int key;
	char *command;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 2)
		XSRETURN_IV(-1);

	key_str = SvPV(ST(0), notused);
	command = SvPV(ST(1), notused);

	if (key_str == NULL || command == NULL)
		XSRETURN_IV(-1);

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_IV(-1);

	bind_add(&screen.binds.main, key, command);
	XSRETURN_IV(0);
}

XS(PORK_unbind) {
	char *key_str;
	int key;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	key_str = SvPV(ST(0), notused);
	if (key_str == NULL)
		XSRETURN_IV(-1);

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_IV(-1);

	key = bind_remove(&screen.binds.main, key);
	XSRETURN_IV(0);
}

XS(PORK_bind_get) {
	char *key_str;
	int key;
	size_t notused;
	struct binding *binding;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_EMPTY;

	key_str = SvPV(ST(0), notused);
	if (key_str == NULL)
		XSRETURN_EMPTY;

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_EMPTY;

	binding = bind_find(&screen.binds.main, key);
	if (binding == NULL)
		XSRETURN_EMPTY;

	XSRETURN_PV(binding->binding);
}

XS(PORK_blist_bind) {
	char *key_str;
	int key;
	char *command;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 2)
		XSRETURN_IV(-1);

	key_str = SvPV(ST(0), notused);
	command = SvPV(ST(1), notused);

	if (key_str == NULL || command == NULL)
		XSRETURN_IV(-1);

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_IV(-1);

	bind_add(&screen.binds.blist, key, command);
	XSRETURN_IV(0);
}

XS(PORK_blist_unbind) {
	char *key_str;
	int key;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	key_str = SvPV(ST(0), notused);
	if (key_str == NULL)
		XSRETURN_IV(-1);

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_IV(-1);

	key = bind_remove(&screen.binds.blist, key);
	XSRETURN_IV(0);
}

XS(PORK_blist_bind_get) {
	char *key_str;
	int key;
	size_t notused;
	struct binding *binding;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_EMPTY;

	key_str = SvPV(ST(0), notused);
	if (key_str == NULL)
		XSRETURN_EMPTY;

	key = bind_get_keycode(key_str);
	if (key == -1)
		XSRETURN_EMPTY;

	binding = bind_find(&screen.binds.blist, key);
	if (binding == NULL)
		XSRETURN_EMPTY;

	XSRETURN_PV(binding->binding);
}

XS(PORK_alias) {
	char *alias;
	char *cmd;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 2)
		XSRETURN_IV(-1);

	alias = SvPV(ST(0), notused);
	cmd = SvPV(ST(1), notused);

	if (alias == NULL || cmd == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(alias_add(&screen.alias_hash, alias, cmd));
}

XS(PORK_alias_get) {
	char *str;
	size_t notused;
	struct alias *alias;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_EMPTY;

	str = SvPV(ST(0), notused);
	if (str == NULL)
		XSRETURN_EMPTY;

	alias = alias_find(&screen.alias_hash, str);
	if (alias == NULL)
		XSRETURN_EMPTY;

	XST_mPV(0, alias->orig);

	if (alias->args != NULL) {
		XST_mPV(1, alias->args);
		XSRETURN(2);
	}

	XSRETURN(1);
}

XS(PORK_unalias) {
	char *alias;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	alias = SvPV(ST(0), notused);
	if (alias == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(alias_remove(&screen.alias_hash, alias));
}

XS(PORK_set_opt) {
	char *var;
	char *val;
	size_t notused;
	int opt;
	dXSARGS;

	(void) cv;

	if (items != 2)
		XSRETURN_IV(-1);

	var = SvPV(ST(0), notused);
	val = SvPV(ST(1), notused);

	if (var == NULL || var == NULL)
		XSRETURN_IV(-1);

	opt = opt_find(var);
	if (opt == -1)
		XSRETURN_IV(-1);

	XSRETURN_IV(opt_set(opt, val));
}

XS(PORK_get_opt) {
	char *val;
	size_t notused;
	char buf[2048];

	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_EMPTY;

	val = SvPV(ST(0), notused);
	if (val == NULL)
		XSRETURN_EMPTY;

	if (opt_get_val(val, buf, sizeof(buf)) == -1)
		XSRETURN_EMPTY;

	XSRETURN_PV(buf);
}

/*
** All 'aim_*' commands take a screen name as their
** first argument.
*/

XS(PORK_aim_send_profile) {
	size_t notused;
	char *profile;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	profile = SvPV(ST(0), notused);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_set_profile(priv, profile));
}

XS(PORK_aim_get_acct_list) {
	dlist_t *cur = screen.acct_list;
	int args = 0;
	dXSARGS;

	(void) cv;

	if (items != 0 || cur == NULL)
		XSRETURN_EMPTY;

	do {
		struct pork_acct *acct = cur->data;

		if (acct->priv->can_connect) {
			XST_mPV(args, acct->username);
			args++;
		}
		cur = cur->next;
	} while (cur != NULL);

	XSRETURN(args);
}

XS(PORK_aim_set_profile) {
	size_t notused;
	char *profile;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	profile = SvPV(ST(0), notused);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	free(priv->profile);
	if (profile != NULL)
		priv->profile = xstrdup(profile);
	else
		priv->profile = NULL;

	XSRETURN_IV(aim_set_profile(priv, profile));
}

XS(PORK_aim_get_profile) {
	size_t notused;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 0)
		XSRETURN_EMPTY;

	if (items == 1) {
		char *username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	if (priv->profile == NULL)
		XSRETURN_EMPTY;

	XSRETURN_PV(priv->profile);
}

XS(PORK_aim_send_im) {
	size_t notused;
	char *dest;
	char *msg;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	dest = SvPV(ST(0), notused);
	msg = SvPV(ST(1), notused);

	if (dest == NULL || msg == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		char *username = SvPV(ST(2), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	return (XSRETURN_IV(aim_send_msg(priv, dest, msg)));
}

XS(PORK_aim_send_im_auto) {
	size_t notused;
	char *dest;
	char *msg;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	dest = SvPV(ST(0), notused);
	msg = SvPV(ST(1), notused);

	if (dest == NULL || msg == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		char *username = SvPV(ST(2), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	return (XSRETURN_IV(aim_send_msg_auto(priv, dest, msg)));
}

XS(PORK_aim_get_buddy_profile) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_whois(priv, target));
}

XS(PORK_aim_get_buddy_away) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_get_away_msg(priv, target));
}

XS(PORK_aim_privacy_mode) {
	size_t notused;
	int privacy_mode;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	privacy_mode = SvIV(ST(0));

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_privacy_mode(priv, privacy_mode));
}

XS(PORK_aim_report_idle) {
	size_t notused;
	int report_idle;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	report_idle = SvIV(ST(0));

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_report_idle(priv, report_idle));
}

XS(PORK_aim_search) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	int ret;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	ret = aim_usersearch_address(&priv->aim_session, priv->bos_conn, target);
	XSRETURN_IV(ret);
}

XS(PORK_aim_set_away) {
	size_t notused;
	char *msg;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	msg = SvPV(ST(0), notused);
	if (msg == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_set_away(priv, msg));
}

XS(PORK_aim_set_idle) {
	size_t notused;
	int seconds;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	seconds = SvIV(ST(0));

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(aim_set_idle(priv, seconds));
}

XS(PORK_aim_warn) {
	size_t notused;
	char *target;
	int anon;
	struct aim_priv *priv;
	int ret;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	anon = SvIV(ST(1));

	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		char *username = SvPV(ST(2), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	if (anon != 0)
		ret = aim_warn(priv, target);
	else
		ret = aim_warn_anon(priv, target);

	XSRETURN_IV(ret);
}

/*
** Wrappers for non-aim /buddy commands.
*/

XS(PORK_buddy_add) {
	size_t notused;
	char *target;
	char *group;
	struct aim_priv *priv;
	struct bgroup *bgroup;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	group = SvPV(ST(1), notused);

	if (target == NULL || group == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		char *username = SvPV(ST(2), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	bgroup = group_find(priv, group);
	if (group == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(buddy_add(priv, target, bgroup, 1) == NULL);
}

XS(PORK_buddy_add_block) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(buddy_add_block(priv, target, 1));
}

XS(PORK_buddy_clear_permit) {
	size_t notused;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		char *username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	buddy_clear_permit(priv);
	XSRETURN_IV(0);
}

XS(PORK_buddy_clear_block) {
	size_t notused;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		char *username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	buddy_clear_block(priv);
	XSRETURN_IV(0);
}

XS(PORK_buddy_add_permit) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(buddy_add_permit(priv, target, 1));
}

XS(PORK_buddy_alias) {
	size_t notused;
	char *target;
	char *alias;
	struct buddy *buddy;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	alias = SvPV(ST(1), notused);

	if (target == NULL || alias == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		char *username = SvPV(ST(2), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	buddy = buddy_find(priv, target);
	if (buddy == NULL)
		XSRETURN_IV(-1);

	buddy_alias(priv, buddy, alias, 0);
	XSRETURN_IV(0);
}

XS(PORK_buddy_get_alias) {
	size_t notused;
	char *target;
	struct buddy *buddy;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_EMPTY;

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_EMPTY;

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	buddy = buddy_find(priv, target);
	if (buddy == NULL)
		XSRETURN_EMPTY;

	XSRETURN_PV(buddy->name);
}

XS(PORK_buddy_remove_block) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(buddy_remove_block(priv, target, 1));
}

XS(PORK_buddy_get_groups) {
	dlist_t *cur;
	int args = 0;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username;

		username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	if (priv->buddy_pref == NULL)
		XSRETURN_EMPTY;

	cur = priv->buddy_pref->group_list;
	while (cur != NULL) {
		struct bgroup *group = cur->data;

		XST_mPV(args, group->name);
		args++;
		cur = cur->next;
	}

	XSRETURN(args);
}

XS(PORK_buddy_get_block) {
	dlist_t *cur;
	int args = 0;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username;

		username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	if (priv->buddy_pref == NULL)
		XSRETURN_EMPTY;

	cur = priv->buddy_pref->block_list;
	while (cur != NULL) {
		XST_mPV(args, cur->data);
		args++;
		cur = cur->next;
	}

	XSRETURN(args);
}

XS(PORK_buddy_get_permit) {
	dlist_t *cur;
	int args = 0;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username;

		username = SvPV(ST(0), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	if (priv->buddy_pref == NULL)
		XSRETURN_EMPTY;

	cur = priv->buddy_pref->permit_list;
	while (cur != NULL) {
		XST_mPV(args, cur->data);
		args++;
		cur = cur->next;
	}

	XSRETURN(args);
}

XS(PORK_buddy_get_group_members) {
	dlist_t *cur;
	char *group;
	struct bgroup *bgroup;
	size_t notused;
	int args = 0;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_EMPTY;

	group = SvPV(ST(0), notused);
	if (group == NULL)
		XSRETURN_EMPTY;

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	bgroup = group_find(priv, group);
	if (bgroup == NULL)
		XSRETURN_EMPTY;

	cur = bgroup->member_list;
	while (cur != NULL) {
		struct buddy *buddy = cur->data;

		XST_mPV(args, buddy->nname);
		args++;
		cur = cur->next;
	}

	XSRETURN(args);
}

XS(PORK_buddy_remove) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(buddy_remove(priv, target, 1));
}

XS(PORK_buddy_remove_group) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(group_remove(priv, target, 1));
}

XS(PORK_buddy_add_group) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	if (group_add(priv, target) == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(0);
}

XS(PORK_buddy_remove_permit) {
	size_t notused;
	char *target;
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	target = SvPV(ST(0), notused);
	if (target == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);
		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(buddy_remove_permit(priv, target, 1));
}

/*
** Scroll commands
*/

XS(PORK_scroll_by) {
	int lines;
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	lines = SvIV(ST(0));

	if (items == 2) {
		u_int32_t refnum = SvIV(ST(1));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_by(imwindow, lines);
	XSRETURN_IV(0);
}

XS(PORK_scroll_down) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_down(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_scroll_end) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_end(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_scroll_page_down) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_page_down(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_scroll_page_up) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_page_up(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_scroll_start) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_start(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_scroll_up) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_scroll_up(imwindow);
	XSRETURN_IV(0);
}

/*
** window commands
*/

XS(PORK_win_find_name) {
	char *str;
	size_t notused;
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	str = SvPV(ST(0), notused);
	if (str == NULL)
		XSRETURN_IV(-1);

	imwindow = imwindow_find_name(str);
	if (imwindow == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(imwindow->refnum);
}

XS(PORK_win_find_target) {
	char *str;
	size_t notused;
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	str = SvPV(ST(0), notused);
	if (str == NULL)
		XSRETURN_IV(-1);

	imwindow = imwindow_find(str);
	if (imwindow == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(imwindow->refnum);
}

XS(PORK_win_bind) {
	char *account;
	size_t notused;
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	account = SvPV(ST(0), notused);
	if (account == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		u_int32_t refnum = SvIV(ST(1));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	XSRETURN_IV(imwindow_bind_acct(imwindow, account));
}

XS(PORK_win_clear) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_clear(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_win_close) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	XSRETURN_IV(screen_close_window(imwindow));
}

XS(PORK_win_erase) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_erase(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_win_next) {
	dXSARGS;

	(void) cv;

	if (items != 0)
		XSRETURN_IV(-1);

	screen_cycle_fwd();
	XSRETURN_IV(0);
}

XS(PORK_win_prev) {
	dXSARGS;

	(void) cv;

	if (items != 0)
		XSRETURN_IV(-1);

	screen_cycle_bak();
	XSRETURN_IV(0);
}

XS(PORK_win_rename) {
	dXSARGS;
	struct imwindow *imwindow;
	char *new_name;
	size_t notused;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	new_name = SvPV(ST(0), notused);
	if (new_name == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		u_int32_t refnum = SvIV(ST(1));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_rename(imwindow, new_name);
	XSRETURN_IV(0);
}

XS(PORK_win_renumber) {
	dXSARGS;
	struct imwindow *imwindow;
	u_int32_t new_refnum;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	new_refnum = SvIV(ST(0));

	if (items == 2) {
		u_int32_t refnum = SvIV(ST(1));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	XSRETURN_IV(screen_renumber(imwindow, new_refnum));
}

XS(PORK_win_get_opt) {
	struct imwindow *imwindow;
	char *val;
	size_t notused;
	char buf[2048];

	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_EMPTY;

	val = SvPV(ST(0), notused);
	if (val == NULL)
		XSRETURN_EMPTY;

	if (items == 2) {
		u_int32_t refnum = SvIV(ST(1));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_EMPTY;
	} else
		imwindow = cur_window();

	if (wopt_get_val(imwindow, val, buf, sizeof(buf)) == -1)
		XSRETURN_EMPTY;

	XSRETURN_PV(buf);
}

XS(PORK_win_set_opt) {
	char *var;
	char *val;
	size_t notused;
	int opt;
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	var = SvPV(ST(0), notused);
	val = SvPV(ST(1), notused);

	if (var == NULL || var == NULL)
		XSRETURN_IV(-1);

	opt = wopt_find(var);
	if (opt == -1)
		XSRETURN_IV(-1);

	if (items == 3) {
		u_int32_t refnum = SvIV(ST(2));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	XSRETURN_IV(wopt_set(imwindow, opt, val));
}

XS(PORK_win_swap) {
	u_int32_t refnum;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	refnum = SvIV(ST(0));
	XSRETURN_IV(screen_goto_window(refnum));
}

/*
** Timer commands.
*/

XS(PORK_timer_add) {
	char *command;
	u_int32_t interval;
	u_int32_t times;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 3)
		XSRETURN_UNDEF;

	interval = SvIV(ST(0));
	times = SvIV(ST(1));
	command = SvPV(ST(2), notused);

	if (command == NULL)
		XSRETURN_UNDEF;

	XSRETURN_IV(timer_add(&screen.timer_list, command, interval, times));
}

XS(PORK_timer_del) {
	char *command;
	size_t notused;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	command = SvPV(ST(0), notused);
	if (command == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(timer_del(&screen.timer_list, command));
}

XS(PORK_timer_del_refnum) {
	u_int32_t refnum;
	dXSARGS;

	(void) cv;

	if (items != 1)
		XSRETURN_IV(-1);

	refnum = SvIV(ST(0));
	XSRETURN_IV(timer_del_refnum(&screen.timer_list, refnum));
}

XS(PORK_timer_purge) {
	dXSARGS;

	(void) items;
	(void) cv;

	timer_destroy(&screen.timer_list);
	XSRETURN_IV(0);
}

/*
** /blist commands.
*/

XS(PORK_blist_collapse) {
	struct aim_priv *priv;
	struct blist *blist;
	size_t notused;
	char *group;
	struct bgroup *bgroup;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	group = SvPV(ST(0), notused);
	if (group == NULL)
		XSRETURN_IV(-1);

	if (items == 2) {
		char *username = SvPV(ST(1), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	bgroup = group_find(priv, group);
	if (bgroup == NULL)
		XSRETURN_IV(-1);

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	blist_collapse_group(blist, bgroup);
	XSRETURN_IV(0);
}

XS(PORK_blist_cursor) {
	struct aim_priv *priv;
	struct blist *blist;
	struct slist_cell *cell;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_EMPTY;

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_EMPTY;

	cell = blist_get_cursor(blist);
	if (cell == NULL)
		XSRETURN_EMPTY;

	if (cell->type == TYPE_FLAT_CELL) {
		struct buddy *buddy = cell->data;

		XST_mPV(0, buddy->nname);
	} else {
		struct bgroup *group = cell->data;

		XST_mPV(0, group->name);
	}

	XST_mIV(1, cell->type);
	XSRETURN(2);
}

XS(PORK_blist_down) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_down(blist));
}

XS(PORK_blist_end) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_end(blist));
}

XS(PORK_blist_hide) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_blist_hide(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_blist_page_down) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_pgdown(blist));
}

XS(PORK_blist_page_up) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_pgup(blist));
}

XS(PORK_blist_refresh) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_refresh(blist));
}

XS(PORK_blist_show) {
	struct imwindow *imwindow;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		u_int32_t refnum = SvIV(ST(0));

		imwindow = imwindow_find_refnum(refnum);
		if (imwindow == NULL)
			XSRETURN_IV(-1);
	} else
		imwindow = cur_window();

	imwindow_blist_show(imwindow);
	XSRETURN_IV(0);
}

XS(PORK_blist_start) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_start(blist));
}

XS(PORK_blist_up) {
	struct aim_priv *priv;
	struct blist *blist;
	dXSARGS;

	(void) cv;

	if (items != 0 && items != 1)
		XSRETURN_IV(-1);

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(blist_cursor_up(blist));
}

XS(PORK_blist_width) {
	struct aim_priv *priv;
	struct blist *blist;
	int new_width;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	new_width = SvIV(ST(0));

	if (items == 2) {
		size_t notused;
		char *username = SvPV(ST(1), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	blist = priv->blist;
	if (blist == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(screen_blist_width(blist, new_width));
}

/*
** Event commands.
*/

XS(PORK_event_add) {
	struct aim_priv *priv;
	char *type;
	char *handler;
	size_t notused;
	dXSARGS;
	u_int32_t refnum;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_UNDEF;

	type = SvPV(ST(0), notused);
	handler = SvPV(ST(1), notused);

	if (type == NULL || handler == NULL)
		XSRETURN_UNDEF;

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_UNDEF;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_UNDEF;
	} else
		priv = cur_window()->priv;

	if (event_add(priv->events, type, handler, &refnum) != 0)
		XSRETURN_UNDEF;

	XSRETURN_IV(refnum);
}

XS(PORK_event_del) {
	struct aim_priv *priv;
	size_t notused;
	char *type;
	char *handler;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	type = SvPV(ST(0), notused);
	handler = SvPV(ST(1), notused);

	if (type == NULL || handler == NULL)
		XSRETURN_IV(-1);

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(event_del_type(priv->events, type, handler));
}

XS(PORK_event_del_type) {
	struct aim_priv *priv;
	size_t notused;
	char *type;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	type = SvPV(ST(0), notused);

	if (items == 2) {
		size_t notused;
		char *username = SvPV(ST(1), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(event_del_type(priv->events, type, NULL));
}

XS(PORK_event_del_refnum) {
	struct aim_priv *priv;
	u_int32_t refnum;
	dXSARGS;

	(void) cv;

	if (items != 1 && items != 2)
		XSRETURN_IV(-1);

	refnum = SvIV(ST(0));

	if (items == 2) {
		size_t notused;
		char *username = SvPV(ST(1), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	XSRETURN_IV(event_del_refnum(priv->events, refnum));
}

XS(PORK_event_purge) {
	struct aim_priv *priv;
	dXSARGS;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	event_purge(priv->events);
	XSRETURN_IV(0);
}

/*
** /chat commands.
*/

XS(PORK_chat_get_list) {
	dXSARGS;
	dlist_t *cur;
	struct aim_priv *priv;
	u_int32_t i = 0;

	(void) cv;

	if (items == 1) {
		size_t notused;
		char *username = SvPV(ST(0), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	for (cur = priv->chat_list ; cur != NULL ; cur = cur->next) {
		struct chatroom *chat;
		char buf[256];

		chat = cur->data;
		snprintf(buf, sizeof(buf), "%s/%d/%s",
			chat->title, chat->exchange, chat->name);

		XST_mPV(i, buf);
		i++;
	}

	XSRETURN(i);
}

XS(PORK_chat_get_window) {
	char *chat_name;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	dlist_t *cur;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	chat_name = SvPV(ST(0), notused);
	if (chat_name == NULL)
		XSRETURN_IV(-1);

	exchange = SvIV(ST(1));

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	cur = aim_find_chat(priv, chat_name, exchange);
	if (cur == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(((struct chatroom *) cur->data)->win->refnum);
}

XS(PORK_chat_get_users) {
	char *chat_name;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	struct chatroom *chat;
	dlist_t *node;
	u_int32_t args = 0;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_EMPTY;

	chat_name = SvPV(ST(0), notused);
	if (chat_name == NULL)
		XSRETURN_EMPTY;

	exchange = SvIV(ST(1));

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_EMPTY;

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_EMPTY;
	} else
		priv = cur_window()->priv;

	node = aim_find_chat(priv, chat_name, exchange);
	if (node == NULL)
		XSRETURN_EMPTY;

	chat = node->data;

	node = chat->user_list;
	while (node != NULL) {
		struct buddy *buddy = node->data;

		XST_mPV(args, buddy->nname);
		args++;

		node = node->next;
	}

	XSRETURN(args);
}

XS(PORK_chat_ignore) {
	char *name;
	char *dest;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	dlist_t *node;
	dXSARGS;

	(void) cv;

	if (items != 3 && items != 4)
		XSRETURN_IV(-1); 

	name = SvPV(ST(0), notused);
	exchange = SvIV(ST(1));
	dest = SvPV(ST(2), notused);

	if (name == NULL || dest == NULL)
		XSRETURN_IV(-1);

	if (items == 4) {
		size_t notused;
		char *username = SvPV(ST(3), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	node = aim_find_chat(priv, name, exchange);
	if (node == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(aim_chat_ignore(dest, node->data));
}

XS(PORK_chat_invite) {
	char *name;
	char *dest;
	char *msg;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	struct chatroom *chat;
	dlist_t *node;
	int ret;
	dXSARGS;

	(void) cv;

	if (items != 4 && items != 5)
		XSRETURN_IV(-1); 

	name = SvPV(ST(0), notused);
	exchange = SvIV(ST(1));
	dest = SvPV(ST(2), notused);
	msg = SvPV(ST(3), notused);

	if (name == NULL || dest == NULL)
		XSRETURN_IV(-1);

	if (items == 5) {
		size_t notused;
		char *username = SvPV(ST(4), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	node = aim_find_chat(priv, name, exchange);
	if (node == NULL)
		XSRETURN_IV(-1);

	chat = node->data;

	ret = aim_chat_invite(&priv->aim_session, priv->bos_conn, dest, msg,
			chat->exchange, chat->name, chat->instance);

	XSRETURN_IV(ret);
}

XS(PORK_chat_join) {
	char *chat_name;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	char buf[128];
	struct imwindow *win;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	chat_name = SvPV(ST(0), notused);
	if (chat_name == NULL)
		XSRETURN_IV(-1);

	exchange = SvIV(ST(1));

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	if (exchange < 4)
		exchange = 4;
	else if (exchange > 16)
		exchange = 16;

	if (aim_find_chat(priv, chat_name, exchange) != NULL) {
		/* We're already a member of this chat room.  Do nothing. */
		XSRETURN_IV(0);
	}

	snprintf(buf, sizeof(buf), "%s/%d", chat_name, exchange);

	win = imwindow_find_chat_target(priv, buf);
	if (win == NULL) {
		win = screen_new_chat_window(priv, buf);
		if (win == NULL)
			XSRETURN_IV(-1);
	}

	XSRETURN_IV(aim_join_chatroom(priv, chat_name, exchange));
}

XS(PORK_chat_leave) {
	char *chat_name;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	dlist_t *cur;
	struct imwindow *chat_win;
	dXSARGS;

	(void) cv;

	if (items != 2 && items != 3)
		XSRETURN_IV(-1);

	chat_name = SvPV(ST(0), notused);
	if (chat_name == NULL)
		XSRETURN_IV(-1);

	exchange = SvIV(ST(1));

	if (items == 3) {
		size_t notused;
		char *username = SvPV(ST(2), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	cur = aim_find_chat(priv, chat_name, exchange);
	if (cur == NULL)
		XSRETURN_IV(-1);

	chat_win = ((struct chatroom *) cur->data)->win;
	if (aim_leave_chatroom(priv, chat_name, exchange) != -1) {
		chat_win->chat = NULL;
		XSRETURN_IV(0);
	}

	XSRETURN_IV(-1);
}

XS(PORK_chat_send) {
	struct aim_priv *priv;
	char *msg;
	char *chat_name;
	int exchange;
	size_t notused;
	dlist_t *cur;
	struct chatroom *chat;
	dXSARGS;

	(void) cv;

	if (items != 3 && items != 4)
		XSRETURN_IV(-1);

	chat_name = SvPV(ST(0), notused);
	exchange = SvIV(ST(1));
	msg = SvPV(ST(2), notused);

	if (chat_name == NULL || msg == NULL)
		XSRETURN_IV(-1);

	if (items == 4) {
		size_t notused;
		char *username = SvPV(ST(3), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	cur = aim_find_chat(priv, chat_name, exchange);
	if (cur == NULL)
		XSRETURN_IV(-1);

	chat = cur->data;
	XSRETURN_IV(aim_chat_send(priv, msg, chat));
}

XS(PORK_chat_unignore) {
	char *name;
	char *dest;
	size_t notused;
	int exchange;
	struct aim_priv *priv;
	dlist_t *node;
	dXSARGS;

	(void) cv;

	if (items != 3 && items != 4)
		XSRETURN_IV(-1); 

	name = SvPV(ST(0), notused);
	exchange = SvIV(ST(1));
	dest = SvPV(ST(2), notused);

	if (name == NULL || dest == NULL)
		XSRETURN_IV(-1);

	if (items == 4) {
		size_t notused;
		char *username = SvPV(ST(3), notused);

		if (username == NULL)
			XSRETURN_IV(-1);

		priv = pork_acct_get_priv(username);
		if (priv == NULL)
			XSRETURN_IV(-1);
	} else
		priv = cur_window()->priv;

	node = aim_find_chat(priv, name, exchange);
	if (node == NULL)
		XSRETURN_IV(-1);

	XSRETURN_IV(aim_chat_unignore(dest, node->data));
}
